// jcdplay.c -- Linux specific low level audio CDROM drive control routines.
// Copyright (c) 1997 by James F. Carter.  1997-12-29, g++
// This is a devolution of WorkMan by Steven Grimm <koreth@hyperion.com> 
//	dated 1995-06-04 (no copyright notice).

/*
Synopsis: 
    jcdplay [-g | -p | -r | -s | -e | -q | -t number] [-d device] [-Q | -v]
        -g      Go (play).  This is the default if no action is specified.  If
		the drive is playing a different track it is moved to the
		selected one, except if the drive is already playing the
		desired track, the track is not started over.
	-p	Pause.  Ignored if the drive is not playing.
	-r	Resume.  Interpreted as -g if drive isn't paused (e.g. stopped).
	-s	Stop.  
	-e	Eject media.  The drive stops first, if playing.
	-q	Query drive state (including track number).
	-t nbr	Play starting at this track.  1 is the default.
        -d dev  The full path name of the CDROM device.  /dev/cdrom (normally a
		symbolic link to the actual device) is the default.  The
		program must have write permission for the device.
		Recommended: make the device mode 660 with a suitable group,
		and make the program setGID to that group.  Or make it 666.
	-Q	Quiet; do not print the mode and track afterward.
	-v	Verbose: show the initial drive state and the actual action.

    When given a command inconsistent with the drive's current state the 
    program transforms it reasonably (e.g. -r is interpreted as -g if the
    drive isn't paused).  No actions (specifically not eject) can be done if
    the CD is mounted.  An attempt to play a data track yields an error 
    message.
*/

/*
 * @(#)plat_linux.c	1.8	5/14/94
 *
 * Linux-specific drive control routines.  Very similar to the Sun module.
 */
static char *ident = "@(#)plat_linux.c	1.8 5/14/94";


#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <ustat.h>		/*ustat is obsolete but statfs doesn't access
					it by device.*/
#include <fcntl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <linux/cdrom.h>
#include <string.h>		/*For memset*/
#include <stdlib.h>		/*for atoi*/


/*
 * Open the CD device and figure out what kind of drive is attached.
 * Returns the file descriptor number, or -1 if no media, or it dies.
 */
int wmcd_open(char* device)
{
	int		fd;
	fd = open(device, 0);
	if (fd < 0 && errno != ENXIO) {
		perror(device);
		exit(1);
	}
	return (fd);
}

/*
 * Returns the number of tracks on the CD, or -1 on failure.
 */
int
gen_get_trackcount(int FD)
{
	struct cdrom_tochdr	hdr;

	if (ioctl(FD, CDROMREADTOCHDR, &hdr))
		return (-1);
	
	return (hdr.cdth_trk1);
}

/*
 * Get the start time and mode (data or audio) of a track.
 */
int
gen_get_trackinfo(
	int FD,			/*Open file descriptor for the drive*/
	int track,		/*Which track you want the data for*/
	int* data,		/*1 if it's a data track, 0 if it's audio*/
	int* startframe)	/*Absolute frame number where it starts*/
{
	struct cdrom_tocentry	entry;

	entry.cdte_track = track;
	entry.cdte_format = CDROM_MSF;

	if (ioctl(FD, CDROMREADTOCENTRY, &entry))
		return (-1);
	
	*startframe =	entry.cdte_addr.msf.minute * 60 * 75 +
			entry.cdte_addr.msf.second * 75 +
			entry.cdte_addr.msf.frame;
	*data = entry.cdte_ctrl & CDROM_DATA_TRACK ? 1 : 0;
	
	return (0);
}

/*
 * Return the number of frames on the CD, or -1 on failure.
 */
int
gen_get_cdlen(int FD)
{
	int		frames, junk;

	gen_get_trackinfo(FD, CDROM_LEADOUT, &junk, &frames);
	return frames;
}


/*
 * Global variables in which the result of drive_status() are returned
 */
enum cd_modes	{
	UNKNOWN = -1,
	BACK = 0, TRACK_DONE = 0,
	PLAYING = 1,
	FORWARD = 2,
	PAUSED=3,
	STOPPED=4,
	EJECTED=5,
	MOUNTED=6
};

char* cd_modes_a[] = 
    {"unknown", "done", "playing", "forward", 
    "paused", "stopped", "ejected", "mounted"};

enum cd_modes mode = UNKNOWN;	/*Current mode of the CD*/
int drive_pos;			/*Position of head relative to start of CD*/
int drive_track;		/*Track currently playing*/
int drive_index;		/*Index (frame offset?) in the track*/
enum cd_modes drive_mode;	/*Mode (PLAYING, EJECTED, etc.) of drive*/

/*
 * Get the current status of the drive: the current play mode, the absolute
 * position from start of disc (in frames), and the current track and index
 * numbers if the CD is playing or paused.  Returns 0 if OK, 1 on failure.
 */
int drive_status(
    int FD)		/*Open file descriptor on the drive*/
{
	struct cdrom_subchnl		sc;
	struct itimerval		old_timer, new_timer;
	int				rstatus;
	struct stat			stbuf;
	struct ustat			ust;

	/* If we can't get status, the CD is ejected, so default to that. */
	enum cd_modes oldmode = drive_mode;
	drive_mode = EJECTED;

	memset(&sc, '\0', sizeof(sc));
	sc.cdsc_format = CDROM_MSF;

	if (ioctl(FD, CDROMSUBCHNL, &sc))
		return (1);
			/*Set the track, even though if drive is not playing
			  the result is garbage.*/
	drive_track = sc.cdsc_trk;
	drive_index = sc.cdsc_ind;
	drive_pos   = sc.cdsc_absaddr.msf.minute * 60 * 75 +
		      sc.cdsc_absaddr.msf.second * 75 +
		      sc.cdsc_absaddr.msf.frame;

			/*Detect if the CD is mounted as a filesystem*/
	if (fstat(FD, &stbuf) == 0 && ustat(stbuf.st_rdev, &ust) == 0) {
		drive_mode = MOUNTED;
	} else switch (sc.cdsc_audiostatus) {
	case CDROM_AUDIO_PLAY:
		drive_mode = PLAYING;
		break;

	case CDROM_AUDIO_PAUSED:
		drive_mode = PAUSED;
		break;
	case CDROM_AUDIO_NO_STATUS:
	case CDROM_AUDIO_INVALID: /**/
		if (oldmode == PLAYING || oldmode == PAUSED)
		{
			drive_mode = PAUSED;
		}
		else
			drive_mode = STOPPED;
		break;

	case CDROM_AUDIO_COMPLETED:
		drive_mode = TRACK_DONE; /* waiting for next track. */
		break;

	default:
		drive_mode = UNKNOWN;
		break;
	}

	return (0);
}

/*
 * Pause the CD.
 */
int gen_pause(int FD)
{
	return (ioctl(FD, CDROMPAUSE));
}

/*
 * Resume playing the CD (assuming it was paused.)
 */
int gen_resume(int FD)
{
	return (ioctl(FD, CDROMRESUME));
}

/*
 * Stop the CD.
 */
int gen_stop(int FD)
{
	return (ioctl(FD, CDROMSTOP));
}

/*
 * Play the CD from one position to another (both in frames.)
 */
int gen_play(			/*Returns 0 if OK, <0 on failure*/
	int FD,			/*Open file descriptor for the drive*/
	int start,		/*Where to start (in absolute frames)*/
	int end)		/*Where to stop*/
{
	struct cdrom_msf		msf;

	msf.cdmsf_min0 = start / (60*75);
	msf.cdmsf_sec0 = (start % (60*75)) / 75;
	msf.cdmsf_frame0 = start % 75;
	msf.cdmsf_min1 = end / (60*75);
	msf.cdmsf_sec1 = (end % (60*75)) / 75;
	msf.cdmsf_frame1 = end % 75;

	if (ioctl(FD, CDROMSTART))
		return (-1);
	if (ioctl(FD, CDROMPLAYMSF, &msf))
		return (-2);
	
	return (0);
}

/*
 * Eject the current CD, if there is one.
 */
int gen_eject(int FD)
{
	struct stat	stbuf;
	struct ustat	ust;

	if (fstat(FD, &stbuf) != 0)
		return (-2);

	/* Is this a mounted filesystem? */
	if (ustat(stbuf.st_rdev, &ust) == 0)
		return (-3);

	if (ioctl(FD, CDROMEJECT))
		return (-1);

	return (0);
}

int play_disc() {
}

enum cd_actions {NOP, GO, PAUSE, RESUME, STOP, EJECT, ERROR};
static char* cd_actions_a[] = 
	{"nothing", "play", "pause", "resume", "stop", "eject", "error"};
	/* 0		1	2	3	4	5	  6 */

static char* device = "/dev/cdrom";
static char* file = NULL;
static char* quiet = "Normal";
static char  action_d[16] = "1";
static char* action = action_d;
static char* track = "-1";

static struct Flags {
    char fbyte;			/*The byte in the flag, e.g. 'v' for -v */
    char** var;			/*Variable to set*/
    char* value;		/*Value to set in it (NULL = next argument)*/
} flags[] = {
    {'q', &action,	"0"},
    {'g', &action,	"1"},
    {'p', &action,	"2"},
    {'r', &action,	"3"},
    {'s', &action,	"4"},
    {'e', &action,	"5"},
    {'t', &track,	NULL},
    {'d', &device,	NULL},
    {'f', &file,	NULL},
    {'Q', &quiet,	"Quiet"},
    {'v', &quiet,	"Verbose"},
    {'\0', NULL,	 NULL} };

    
int main (int argn, char* argv[]) {
    struct Flags* f;
    int bad = 0;
    int FD;
    int start, end, data, action_x, track_req;
		/* xa_xlate[action][state] gives modified action as a function
		    of the CD's state. */
    static char* xa_xlate[] = {
	"00000000",			/*nop*/
	"61003166",			/*play*/
	"60220066",			/*pause*/
	"61003166",			/*resume*/
	"60440066",			/*stop*/
	"65555556",			/*eject*/
	"66666666" };			/*error*/
	
		/* Interpret command line arguments */
    while (*(++argv) != NULL) {
	if (argv[0][0] != '-') {
	    bad |= 1;
	} else {
	    for (f = flags; f->fbyte != '\0' && f->fbyte != argv[0][1]; ++f)
		/*no-op*/ ; 
	    if (f->fbyte == '\0') { 
		bad |= 1; 
	    } else if (f->value) {
		*(f->var) = f->value;
	    } else {
		*(f->var) = *++argv;
	    }
	}
	if (bad & 1) {
	    bad = 2;
	    fprintf(stderr, "Unrecognized switch %s\n", *argv);
	}
    }
    if (bad) {
	fputs(
"Usage: jcdplay [-g | -p | -r | -s | -e | -q | -t number] [-d device] [-Q | -v]
    Translation: -go, -pause, -resume, -stop, -eject, -query, -track, -Quiet\n", stderr);
	exit(4);
    }
		/* Open the drive. */
    FD = wmcd_open(device);
    if (FD < 0) { exit(4); }
		/* What is the status of the drive? */
		/* Translate (drive status x requested action) into the real
		    action to be taken. */
    drive_status(FD);
    track_req = atoi(track);
    action_x = xa_xlate[((int)*action) - (int)'0'][(int)drive_mode+1];
		/* If no track was specified, assume track 1.  If it was
		    specified and the drive is playing a different track,
		    override the NOP action; play the specified track. */
    if (track_req > 0) {
	if (*action == '1' && action_x == '0' && drive_track != track_req)
	    action_x = '1';
    } else {
	track_req = 1;
    }
    if (*quiet == 'V') {
	printf("Action to be performed: %s  Drive status: %s\n",
	    cd_actions_a[action_x - '0'], cd_modes_a[drive_mode+1]);
    }
		/* Perform the requested action */
    switch (action_x) {
	case '0':			/* Do nothing */
	    break;
	case '1':			/* -g or -t <track>: play that track */
	    bad = gen_get_trackinfo(FD, track_req, &data, &start);
	    end = gen_get_cdlen(FD);
	    if (start >= end) { 
		fprintf(stderr, "Track %d (at %d) is after the end of the CD (at %d)\n",
		    track_req, start, end);
		bad = 1;
	    }
	    if (data) {
		fprintf(stderr, "Track %d contains data.  Use -t <number> to pick an audio track.\n", track_req);
		bad = 1;
	    }
	    if (!bad) { bad = gen_play(FD, start, end); }
	    break;
	case '2': bad = gen_pause(FD);  break;
	case '3': bad = gen_resume(FD); break;
	case '4': bad = gen_stop(FD);   break;
	case '5': bad = gen_eject(FD);  break;
	default:
	    fprintf(stderr, "Cannot %s when the CD is %s\n", 
		cd_actions_a[*action - '0'], cd_modes_a[drive_mode+1]);
	    bad = 1;
    };
		/* Report the current drive status */
    if (bad || *quiet != 'Q') {
	drive_status(FD);
	printf("Mode:	%s	(%s)	Track:	%d	Index:	%d\n",
	    cd_modes_a[1+(int)drive_mode], 
	    (bad ? "failed" : "OK"), 
	    drive_track, drive_index);
    }
    close(FD);
    exit(bad);
}
