// jxcd.cc -- Linux specific low level audio CDROM drive control routines.
// Copyright (c) 1997 by James F. Carter.  1997-12-29, g++
// This is a devolution of WorkMan by Steven Grimm <koreth@hyperion.com> 
//	dated 1995-06-04 (no copyright notice).  See jxcd (tcl script) for
//	freeware license statement.

//	The CDROM control routines have been converted to be callable from tcl.
//	A tk GUI has been provided as a separate script which is executed by
//	this program as the interpreter.

/*
 * Synopsis: Create a tcl/tk script beginning #!/this/program -file
 * Within the script you can do [cdrom command args] where the available
 * commands are:
 *	open $device	Open the cdrom.  The $device filename is optional; the
 *			default is /dev/cdrom.
 *	close		Close the cdrom.  Normally done automatically by the
 *			destructor.
 *	cdlen		Returns the number of frames on the media.
 *	trackcount	Returns the number of tracks on the media.
 *	trackinfo $trk	Returns info about that track (1..max), as a list:
 *			{min sec frame absframe data}, where the position is
 *			relative to the start of the disc, and $data is 1 for
 *			a data track or 0 for audio.
 *	status		Returns the drive status as a list:
 *			{min sec frame absframe track status}, where the msf
 *			position is relative to the start of the track, absframe
 *			is relative to the start of the disc, tracks run from
 *			1 to max, and the status is: unknown, done, playing, 
 *			forward, paused, stopped, ejected, mounted.
 *	pause		Pause the drive (track and offset stay set up).
 *	resume		Resume playing after pause.
 *	stop		Stop the drive (track and offset are lost).
 *	eject		Eject the media.
 *	play $s $e	Play the disc starting at frame $s (relative to the
 *			beginning of the disc), and ending at $e (normally the
 *			result of [cdrom cdlen]).
 *	nothing		Does nothing (to simplify calling program's logic).
 */

/*
 * 
 */

/*
 * Indicia from the WorkMan file that this was derived from:
 * @(#)plat_linux.c	1.8	5/14/94
 * Linux-specific drive control routines.  Very similar to the Sun module.
static char *ident = "@(#)plat_linux.c	1.8 5/14/94";
 */

#include "jxcd.h"		/*Includes Tcl++.h*/

#include <stdio.h>		/*For NULL and misc. debug printf's*/
#include <errno.h>		/*For errno and sys_errlist*/
#include <unistd.h>		/*for close*/
#include <sys/types.h>		/*for arguments to open, and size_t?*/
#include <ustat.h>		/*ustat is obsolete but statfs doesn't access
					it by device.*/
#include <fcntl.h>		/*for open*/
#include <sys/ioctl.h>		/*for ioctl*/
#include <sys/stat.h>		/*for stat and ustat*/
#include <linux/cdrom.h>	/*for CDROM struct and constants*/
#include <string.h>		/*For memset*/
#include <stdlib.h>		/*for atoi*/

#include <tk.h>

// Static data members

/*static*/ const CdromCmd::CMDTAB CdromCmd::cmdtab[] = {
    { "open",		"C",	&CdromCmd::wmcd_open	},
    { "close",		"",	&CdromCmd::wmcd_close	},
    { "cdlen",		"",	&CdromCmd::cdlen	},
    { "trackcount",	"",	&CdromCmd::trackcount	},
    { "trackinfo",	"i",	&CdromCmd::trackinfo	},
    { "status",		"",	&CdromCmd::status	},
    { "pause",		"",	&CdromCmd::pause	},
    { "resume",		"",	&CdromCmd::resume	},
    { "stop",		"",	&CdromCmd::stop		},
    { "eject",		"",	&CdromCmd::eject	},
    { "play",		"ii",	&CdromCmd::play		},
    { "nothing",	"",	&CdromCmd::nothing	},
};


/*static*/ const char* CdromCmd::cd_actions_a[] = 
    {"nothing",	"play",	"pause", "resume", "stop", "eject", "error"};
    /* 0	  1	    2	    3	    4		5	6 */

/*static*/ const char* CdromCmd::cd_modes_a[] = 
    {"unknown", "done", "playing", "forward", 
    "paused", "stopped", "ejected", "mounted"};
    /* 0,4	1,5	    2,6		3,7 */

/*static*/ TclDString CdromCmd::device("/dev/cdrom");

/* Package general initialization */
CdromCmd::CdromCmd(TclCopy& tcl) : TclCommand(tcl, "cdrom"), FD(-1) {
    // Not much programmatic initialization for this class
}

/* Main command executor, parses arguments and calls individual subcommands.*/
int CdromCmd::Cmd(int argc, char* argv[]) {
    int cx, ax, res, i;
    int* iap;
    char* q;
    if (argc < 2) {
	SetResult("Subcode and args required");
	return TCL_ERROR;
    }
    for (cx = 0; cx < sizeof(cmdtab)/sizeof(cmdtab[0]); ++cx) {
	if (!strcmp(argv[1], cmdtab[cx].name))
	    break;
    }
    if (cx >= sizeof(cmdtab)/sizeof(cmdtab[0])) {
	AppendResult("Unrecognized subcode ", argv[1], 
	    "; expecting:", NULL);
	for (cx = 0; cx < sizeof(cmdtab)/sizeof(cmdtab[0]); ++cx) {
	    AppendElement(cmdtab[cx].name);
	}
	return TCL_ERROR;
    }
		/*Parse the args given to the command.*/
    q = cmdtab[cx].args;
    i = strlen(q);
    if (!(argc-2 == i || (q[i-1] == 'C' && argc-1 == i))) {
	Resprintf("Wrong # args, got %d args, cdrom %s needs %d args", 
	    argc-2, cmdtab[cx].name, i);
	return TCL_ERROR;
    }
    ax = 0;
    res = TCL_OK;
    for ( ; *q; ++q, ++ax) {
	cargs[ax] = argv[ax+2];
	iap = &(iargs[ax]);
	switch (*q) {
	    case 'i':
		res = Get(argv[ax+2], iap);
		break;
	    case 'd':
		res = Get(argv[ax+2], &(dargs[ax]));
	    /*case 'c' and 'C' are handled automatically, storing in cargs[ax]*/
	}
	if (res != TCL_OK) {
	    return res;
	}
    }
    return (this->*(cmdtab[cx].fcn))();
}

/*
 * Open the CD device and figure out what kind of drive is attached.
 * [cdrom open $device] (device is optional, default is /dev/cdrom). 
 * Returns 0 if opened or 1 if no media or does an error exit on any 
 * other I/O error.
 */
int CdromCmd::wmcd_open() {
    int res = TCL_OK;
    if (cargs[0] != NULL) device = cargs[0];
    FD = open((char*)device, 0);
    if (FD >= 0) {
	SetResult("0");			//It worked
    } else if (errno == ENXIO) {
	SetResult("1");			//No media
    } else {
	Resprintf("%s: %s (%d)", (const char*)device, sys_errlist[errno], errno);
	res = TCL_ERROR;
    }
    return res;
}

/*
 * Close the device.
 */
int CdromCmd::wmcd_close() {
    if (FD >= 0)
	close(FD);
    FD = -1;			//It isn't open any more
    return TCL_OK;
}

/*
 * Return the number of frames on the CD, or -1 on failure.
 */
int CdromCmd::cdlen() {
    Resprintf("%d", itrackinfo(CDROM_LEADOUT));
    return TCL_OK;
}


/*
 * Returns the number of tracks on the CD, or -1 on failure.
 */
int CdromCmd::trackcount() {
    struct cdrom_tochdr	hdr;
    int tracks = -1;
    if (!ioctl(FD, CDROMREADTOCHDR, &hdr)) {
	tracks = hdr.cdth_trk1;
    }
    Resprintf("%d", tracks);
    return TCL_OK;
}

/*
 * Get the start time and mode (data or audio) of a track.
 * [cdrom trackinfo $track] returns {minute second frame absframe data} where:
 *	minute	Number of minutes from start of disc to start of track.
 *	second	Seconds after that minute
 *	frame	Frames after that second
 *	absframe  Frames after the start of the disc, where the track starts
 *	data	1 if it's a data track, 0 if audio, -1 if failed.
 * Special case: track -1 refers to the end of the disc.
 */
int CdromCmd::trackinfo() {
    itrackinfo(iargs[0]);
    return TCL_OK;
}
int CdromCmd::itrackinfo(int track) {
    struct cdrom_tocentry	entry;
    
    if (track == -1) track = CDROM_LEADOUT;
    memset(&entry, '\0', sizeof(entry));
    entry.cdte_track = track;
    entry.cdte_format = CDROM_MSF;
    
    int res = ioctl(FD, CDROMREADTOCENTRY, &entry);
    int datatrack = res ? -1 :
	    ((entry.cdte_ctrl & CDROM_DATA_TRACK) ? 1 : 0);
    
    int absframe =  entry.cdte_addr.msf.frame  + CD_FRAMES * (
		    entry.cdte_addr.msf.second + CD_SECS * 
		    entry.cdte_addr.msf.minute);
    Resprintf("%d %d %d %d %d",
	(int)(entry.cdte_addr.msf.minute),
	(int)(entry.cdte_addr.msf.second),
	(int)(entry.cdte_addr.msf.frame),
	absframe, datatrack);
    return absframe;
}

/*
 * Get the current status of the drive.  [cdrom status] returns a list of:
 *	minute	Number of minutes from start of disc to current play position
 *	second	Seconds after that minute
 *	frame	Frames after that second
 *	absframe  Frames after the start of the disc to play position
 *	track	Which track we're on
 *	status	The drive status (playing, paused, etc.)
 * The current status is also saved in the data structure.
 */
int CdromCmd::status() {
    struct cdrom_subchnl sc;
    int	 rstatus;
    struct stat	stbuf;
    struct ustat ust;
    cd_modes oldmode = mode;

		// Find out where the drive is playing.  If not playing or no
		// media or (etc.) the result will be garbage; what me worry?
    memset(&sc, '\0', sizeof(sc));
    sc.cdsc_format = CDROM_MSF;
    int res = ioctl(FD, CDROMSUBCHNL, &sc);
    int drive_pos = sc.cdsc_absaddr.msf.frame  + CD_FRAMES * (
		    sc.cdsc_absaddr.msf.second + CD_SECS *
		    sc.cdsc_absaddr.msf.minute);

			/*Detect if the CD is mounted as a filesystem*/
			/*WARNING, Linux 2.0.29 docs say ustat is provided
			    for backward compatibility only and statfs is
			    recommended.  In freebsd you can do 
			    fstatfs(stbuf.st_rdev).  ustat is used in HPUX,
			    linux, Sony-news, OSF1, Sun (SunOS-4?) and Ultrix.
			    bsd386 and SVR4 can't detect if the CD is mounted.*/
    if (res) {
	mode = EJECTED;
    } else if (fstat(FD, &stbuf) == 0 && ustat(stbuf.st_rdev, &ust) == 0) {
	mode = MOUNTED;
    } else switch (sc.cdsc_audiostatus) {
	case CDROM_AUDIO_PLAY:
	    mode = PLAYING;
	    break;
	case CDROM_AUDIO_PAUSED:
	    mode = PAUSED;
	    break;
	case CDROM_AUDIO_NO_STATUS:
	case CDROM_AUDIO_INVALID: /**/
//	    mode = (oldmode == PLAYING || oldmode == PAUSED) ? PAUSED : STOPPED;
	    mode = STOPPED;
	    break;
	case CDROM_AUDIO_COMPLETED:
	    mode = TRACK_DONE; /* waiting for next track. */
	    break;
	default:
	    mode = UNKNOWN;
	    break;
    }

    Resprintf("%d %d %d %d %d %s", 
	(int)(sc.cdsc_absaddr.msf.minute),
	(int)(sc.cdsc_absaddr.msf.second),
	(int)(sc.cdsc_absaddr.msf.frame),
	drive_pos, 
	(int)(sc.cdsc_trk),
	cd_modes_a[mode]);
    return TCL_OK;
}

/*
 * Generically issue an action IOCTL.
 */
int CdromCmd::aioctl(int code, void* data) {
    int res = ioctl(FD, code, data);
    if (res) {
	res = TCL_ERROR;
	Resprintf("%s: %s (%d)", (const char*)device, sys_errlist[errno], errno);
    } else {
	res = TCL_OK;
    }
    return res;
}

/*
 * Pause the CD.
 */
int CdromCmd::pause() {
	return aioctl(CDROMPAUSE, NULL);
}

/*
 * Resume playing the CD (assuming it was paused.)
 */
int CdromCmd::resume() {
	return aioctl(CDROMRESUME, NULL);
}

/*
 * Stop the CD.
 */
int CdromCmd::stop() {
	return aioctl(CDROMSTOP, NULL);
}

/*
 * Eject the current CD, if there is one.
 */
int CdromCmd::eject() {
	return aioctl(CDROMEJECT, NULL);
}

/*
 * Play the CD from one position to another.  [cdrom play $start $end] 
 * makes it play between the indicated limits, which are in absolute frames
 * relative to the start of the disc.
 */
int CdromCmd::play() {
	int res;
	struct cdrom_msf		msf;
	
	msf.cdmsf_min0 = iargs[0] / (CD_SECS*CD_FRAMES);
	msf.cdmsf_sec0 = (iargs[0] % (CD_SECS*CD_FRAMES)) / CD_FRAMES;
	msf.cdmsf_frame0 = iargs[0] % CD_FRAMES;
	msf.cdmsf_min1 = iargs[1] / (CD_SECS*CD_FRAMES);
	msf.cdmsf_sec1 = (iargs[1] % (CD_SECS*CD_FRAMES)) / CD_FRAMES;
	msf.cdmsf_frame1 = iargs[1] % CD_FRAMES;

	res = aioctl(CDROMSTART, NULL);
	if (res == TCL_OK)
	    res = aioctl(CDROMPLAYMSF, &msf);
	
	return res;
}

/*
 * Subroutine to do nothing, to simplify a caller's logic.
 */
int CdromCmd::nothing() {
    return TCL_OK;
}

void bitmapsetup(Tcl_Interp* interp) {
		//The six bitmaps
#define eject_width 23
#define eject_height 17
static char eject_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x3e, 0x00,
   0x00, 0x7f, 0x00, 0x80, 0xff, 0x00, 0xc0, 0xff, 0x01, 0xe0, 0xff, 0x03,
   0xf0, 0xff, 0x07, 0xf8, 0xff, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xf8, 0xff, 0x0f, 0xf8, 0xff, 0x0f, 0xf8, 0xff, 0x0f, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};
#define stop_width 23
#define stop_height 17
static char stop_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f,
   0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f,
   0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f,
   0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0xe0, 0xff, 0x0f, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};
#define pause_width 23
#define pause_height 17
static char pause_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03,
   0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03,
   0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03,
   0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0xc0, 0xc3, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};
#define rew_width 23
#define rew_height 17
static char rew_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x0c, 0x00, 0x07, 0x0e,
   0x80, 0x07, 0x0f, 0xc0, 0x87, 0x0f, 0xe0, 0xc7, 0x0f, 0xf0, 0xe7, 0x0f,
   0xf8, 0xff, 0x0f, 0xf0, 0xe7, 0x0f, 0xe0, 0xc7, 0x0f, 0xc0, 0x87, 0x0f,
   0x80, 0x07, 0x0f, 0x00, 0x07, 0x0e, 0x00, 0x06, 0x0c, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};
#define play_width 23
#define play_height 17
static char play_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x00, 0x00, 0x3f, 0x00,
   0x00, 0x7f, 0x00, 0x00, 0xff, 0x00, 0x00, 0xff, 0x01, 0x00, 0xff, 0x03,
   0x00, 0xff, 0x07, 0x00, 0xff, 0x03, 0x00, 0xff, 0x01, 0x00, 0xff, 0x00,
   0x00, 0x7f, 0x00, 0x00, 0x3f, 0x00, 0x00, 0x1f, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};
#define ff_width 23
#define ff_height 17
static char ff_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x60, 0x00, 0x70, 0xe0, 0x00,
   0xf0, 0xe0, 0x01, 0xf0, 0xe1, 0x03, 0xf0, 0xe3, 0x07, 0xf0, 0xe7, 0x0f,
   0xf0, 0xff, 0x1f, 0xf0, 0xe7, 0x0f, 0xf0, 0xe3, 0x07, 0xf0, 0xe1, 0x03,
   0xf0, 0xe0, 0x01, 0x70, 0xe0, 0x00, 0x30, 0x60, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00};

    struct BITTAB {
	char* name;
	char* bits;
	int width;
	int height;
    };
    static BITTAB bittab[] = {
	{"eject",	eject_bits,	eject_width,	eject_height	},
	{"ff",		ff_bits,	ff_width,	ff_height	},
	{"pause",	pause_bits,	pause_width,	pause_height	},
	{"play",	play_bits,	play_width,	play_height	},
	{"rew",		rew_bits,	rew_width,	rew_height	},
	{"stop",	stop_bits,	stop_width,	stop_height	},
	{NULL,		NULL,		0,		0		}
    };
    BITTAB* b;
    for (b = bittab; b->name != NULL; ++b) {
	Tk_DefineBitmap(interp, Tk_GetUid(b->name), 
	    b->bits, b->width, b->height);
    }
}

// Main program

int Tcl_AppInit(Tcl_Interp* interp) {
    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
        return TCL_ERROR;
    }
    static TclCopy* mytcl = new TclCopy(interp);
    static CdromCmd* cdrom;
    cdrom = new CdromCmd(*mytcl);
    
    bitmapsetup(interp);
    
    return ((cdrom == NULL) ? TCL_ERROR : TCL_OK);
}

/* In tk8.3, a macro is used to define Tk_Main. */
#ifndef Tk_Main
extern void Tk_Main(int argc, char** argv, Tcl_AppInitProc *appInitProc);
#endif

int main(int argc, char** argv) {
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;
}

